@@ -16,7 +16,7 @@
 # 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 #
 '''
-Hardware Abstraction Layer (hal) related classes.
+Hardware Abstraction Layer (HAL) related classes.
 '''
 
 import dbus
@@ -25,7 +25,8 @@ import gtk.gdk
 import logging
 import os, os.path
 
-from ConfigParser import SafeConfigParser
+from ConfigParser          import SafeConfigParser
+from gettext               import gettext as _
 from gnome_lirc_properties import lirc
 
 HAL_SERVICE = 'org.freedesktop.Hal'
@@ -68,9 +69,7 @@ class HalDevice(object):
 
     def lookup_parent(self):
         '''Find the parent device of this device.'''
-        udi = self['info.parent']
-        # TODO: pychecker says "gnome_lirc_properties/hardware.py:72: Invalid arguments to (__init__), got 2, expected 3"
-        return HalDevice(self.__bus, udi)
+        return HalDevice(self.__bus, self.__hal, self['info.parent'])
 
     def find_children(self):
         '''Lists all child devices of this device.'''
@@ -90,6 +89,10 @@ class HalDevice(object):
         class_root = os.path.join(os.sep, 'sys', 'class', device_class)
         sysfs_path = self['linux.sysfs_path']
 
+        # the class root does not exist when no such driver is loaded:
+        if not os.path.isdir(class_root):
+            return None
+
         for device in os.listdir(class_root):
             try:
                 # resolve the device link found in the sysfs folder:
@@ -128,6 +131,7 @@ class HalDevice(object):
 
     def find_device_node(self, kernel_module, lirc_driver):
         '''Finds the device node associated with this device.'''
+
         if 'usbhid' == kernel_module:
             return self.find_device_by_class('usb')
         if 'default' == lirc_driver:
@@ -163,7 +167,15 @@ class HardwareManager(gobject.GObject):
             tuple()),
         'receiver-found': (
             gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE,
-            (gobject.TYPE_PYOBJECT, gobject.TYPE_STRING, gobject.TYPE_STRING)),
+            (gobject.TYPE_PYOBJECT, gobject.TYPE_STRING,
+             gobject.TYPE_STRING)),
+
+        'receiver-added': (
+            gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE,
+            (gobject.TYPE_PYOBJECT, )),
+        'receiver-removed': (
+            gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE,
+            (gobject.TYPE_PYOBJECT, ))
     }
 
     def __init__(self, hardware_db):
@@ -172,7 +184,21 @@ class HardwareManager(gobject.GObject):
         assert(isinstance(hardware_db, HardwareDatabase))
 
         super(HardwareManager, self).__init__()
+
+        self.__devinput_receivers = dict()
         self.__search_canceled = False
+        self.__device_count = 0.0
+        self.__progress = 0.0
+
+        self.__bus = dbus.SystemBus()
+        self.__hal = self.__bus.get_object(HAL_SERVICE, HAL_MANAGER_PATH)
+        self.__hal = dbus.Interface(self.__hal, HAL_MANAGER_IFACE)
+
+        self.__hal.connect_to_signal('DeviceAdded', self._on_device_added)
+        self.__hal.connect_to_signal('DeviceRemoved', self._on_device_removed)
+
+        for udi in self.__hal.FindDeviceByCapability('input.keyboard'):
+            self._on_device_added(udi)
 
         # reading hardware database, and reording as
         # dictionary with (product-id, vendor-id) keys
@@ -183,25 +209,82 @@ class HardwareManager(gobject.GObject):
             properties = dict(hardware_db.items(section))
 
             receiver = lirc.Receiver(vendor, product, **properties)
-            key = (receiver.vendor_id, receiver.product_id)
+
+            if receiver.vendor_id:
+                key = (receiver.vendor_id, receiver.product_id)
+
+            elif receiver.kernel_module:
+                key = receiver.kernel_module
+
+            else:
+                key = receiver.lirc_driver
+
             self.__receivers[key] = receiver
 
+    def __get_devinput_receivers(self):
+        '''List all currently known Input Device Layer receiver.'''
+        return self.__devinput_receivers
+
+    def _on_device_added(self, udi, sender=None):
+        '''Handle addition of hot-plugable devices.'''
+
+        device = self.lookup_device(udi)
+
+        if 'input.keyboard' in device.get('info.capabilities', []):
+            product_name = str(device['info.product'])
+            device_node = str(device['input.device'])
+
+            properties = {
+                'compatible-remotes': 'linux-input-layer',
+                'lirc-driver': 'devinput',
+                'device': device_node,
+                'udi': udi,
+            }
+
+            receiver = lirc.Receiver(_('Linux Input Device'),
+                                     product_name, **properties)
+
+            self.__devinput_receivers[udi] = receiver
+            self.emit('receiver-added', receiver)
+
+    def _on_device_removed(self, udi, sender=None):
+        '''Handle removal of hot-plugable devices.'''
+
+        receiver = self.__devinput_receivers.pop(udi, None)
+
+        if receiver is not None:
+            self.emit('receiver-removed', receiver)
+
     def resolve_device_nodes(self, nodes):
         '''Resolve the requested device nodes.'''
-        device_nodes, hal, bus = list(), None, None
+        device_nodes = list()
 
         for name in nodes:
             # Use HAL manager to find device nodes,
             # when name starts with "hal-capability:"
             if name.startswith('hal-capability:'):
                 capability = name.split(':')[1]
+                device_property = '%s.device' % capability.split('.')[0]
 
-                if not hal:
-                    hal, bus = self._create_hal_manager()
+                # Find devices with requested capability:
+
+                devices = [
+                    self.lookup_device(udi)
+                    for udi in self.__hal.FindDeviceByCapability(capability)]
+
+                # Extract device node and product name for matching devices:
+
+                devices = [(
+                    device['info.product'],
+                    device.get(device_property))
+                    for device in devices]
+
+                # Throw away devices without device node:
 
                 device_nodes += [
-                    str(HalDevice(bus, hal, udi)['serial.device'])
-                    for udi in hal.FindDeviceByCapability(capability)]
+                    (str(name), str(device))
+                    for name, device in devices
+                    if device is not None]
 
             elif name.startswith('numeric:'):
                 device_nodes.append(name)
@@ -230,34 +313,21 @@ class HardwareManager(gobject.GObject):
 
         return [label] + limits
 
-    @staticmethod
-    def _create_hal_manager():
-        '''Retreive HAL manager instance from DBus'''
+    def __report_search_progress(self, product_name):
+        '''Reports progress when searching supported devices.'''
 
-        bus = dbus.SystemBus()
-        hal = bus.get_object(HAL_SERVICE, HAL_MANAGER_PATH)
-        hal = dbus.Interface(hal, HAL_MANAGER_IFACE)
+        self._search_progress(self.__progress / self.__device_count,
+                              product_name)
 
-        return hal, bus
+        logging.info('%d/%d: %s', self.__progress,
+                     self.__device_count, product_name)
 
-    def search_receivers(self):
-        '''
-        Search for supported IR receivers. This search actually happens
-        in a separate thread and can be aborted with the cancel() method.
-        Connect to the signals of this object, to monitor search progress.
+        self.__progress += 1.0
 
-        TODO: Currently the search happens in the main thread, as DBus' Python
-        bindings still cause random dead-locks when used with threads. Despite
-        claiming being thread-safe.
-        '''
+    def __find_usb_receivers(self, usb_devices):
+        '''Find IR receivers connected via USB bus.'''
 
-        self.__search_canceled = False
-
-        # retreive list of USB devices from HAL
-        hal, bus = self._create_hal_manager()
-        usb_devices = hal.FindDeviceStringMatch('info.subsystem', 'usb_device')
-
-        for i in range(len(usb_devices)):
+        for udi in usb_devices:
             # provide cancelation point:
             if self.__search_canceled:
                 break
@@ -268,16 +338,19 @@ class HardwareManager(gobject.GObject):
                     return
 
             # lookup the current device:
-            device = HalDevice(bus, hal, usb_devices[i])
-
-            # report search progress:
-            self._search_progress(float(i)/float(len(usb_devices)),
-                                  '%s %s', device['info.vendor'],
-                                           device['info.product'])
+            device = self.lookup_device(udi)
 
             # lookup the USB devices in our hardware database:
-            vendor_id = device['usb_device.vendor_id']
-            product_id = device['usb_device.product_id']
+            vendor_id = device.get('usb_device.vendor_id')
+            product_id = device.get('usb_device.product_id')
+
+            # skip USB devices that do not get sufficient power:
+            if vendor_id is None or product_id is None:
+                continue
+
+            # report search progress:
+            self.__report_search_progress('%s %s' % (device['info.vendor'],
+                                                     device['info.product']))
 
             # Skip devices without vendor id. Linux for instance
             # doesn't assign a vendor id to USB host controllers.
@@ -311,6 +384,78 @@ class HardwareManager(gobject.GObject):
 
             self._receiver_found(receiver, device.udi, device_node)
 
+    def __find_input_layer_receivers(self, input_devices):
+        '''Find IR receivers implementing the Linux Input Layer interface.'''
+
+        for device in input_devices:
+            # provide cancelation point:
+            if self.__search_canceled:
+                break
+
+            # process GTK+ events:
+            while gtk.events_pending():
+                if gtk.main_iteration():
+                    return
+
+            # lookup the current device:
+            device = self.lookup_device(device)
+            product_name = device.get('info.product')
+            device_node = device.get('input.device')
+
+            # report search progress:
+            self.__report_search_progress(product_name)
+
+            # skip input device that do not provide sufficient information:
+            if product_name is None or device_node is None:
+                continue
+
+            # skip input devices that have the word "keyboard" in their product name:
+            if product_name.lower().find('keyboard') >= 0:
+                continue
+
+            # report findings:
+            receiver = self.devinput_receivers[device.udi]
+            self._receiver_found(receiver, device.udi, device_node)
+
+    def find_instance(self, receiver):
+        '''Finds the device node associated with this receiver.'''
+
+        for udi in self.__hal.FindDeviceStringMatch('info.subsystem', 'usb_device'):
+            device = self.lookup_device(udi)
+
+            if (device['usb_device.vendor_id'] == receiver.vendor_id and
+                device['usb_device.product_id'] == receiver.product_id):
+                return device.find_device_node(receiver.kernel_module,
+                                               receiver.lirc_driver)
+
+        return None
+
+    def lookup_device(self, udi):
+        '''Looks up the HAL device for that UDI.'''
+        return HalDevice(self.__bus, self.__hal, udi)
+
+    def search_receivers(self):
+        '''
+        Search for supported IR receivers. This search actually happens
+        in a separate thread and can be aborted with the cancel() method.
+        Connect to the signals of this object, to monitor search progress.
+
+        TODO: Currently the search happens in the main thread, as DBus' Python
+        bindings still cause random dead-locks when used with threads. Despite
+        claiming being thread-safe.
+        '''
+
+        # retreive list of USB devices from HAL
+        usb_devices = self.__hal.FindDeviceStringMatch('info.subsystem', 'usb_device')
+        input_devices = self.__hal.FindDeviceByCapability('input.keyboard')
+
+        self.__search_canceled = False
+        self.__device_count = float(len(usb_devices) + len(input_devices))
+        self.__progress = 0.0
+
+        self.__find_usb_receivers(usb_devices)
+        self.__find_input_layer_receivers(input_devices)
+
         self._search_finished()
 
     def _search_progress(self, progress, message, *args):
@@ -335,3 +480,5 @@ class HardwareManager(gobject.GObject):
         '''Abort current search process.'''
 
         self.__search_canceled = True
+
+    devinput_receivers = property(__get_devinput_receivers)
